# -*- coding: utf-8 -*-
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Glance Release Notes documentation build configuration file, created by
# sphinx-quickstart on Tue Nov  3 17:40:50 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

# -- General configuration ------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'openstackdocstheme',
    'reno.sphinxext',
]

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'osc_placement Release Notes'
copyright = '2016, OpenStack Foundation'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
# The full version, including alpha/beta/rc tags.
release = ''
# The short X.Y version.
version = ''

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'openstackdocs'

# openstackdocstheme options
openstackdocs_repo_name = 'openstack/osc-placement'
openstackdocs_auto_name = False
openstackdocs_use_storyboard = True



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from osc_lib.command import command
from osc_lib import utils

from osc_placement import version


BASE_URL = '/resource_providers/{uuid}/usages'
USAGES_URL = '/usages'
FIELDS = ('resource_class', 'usage')


class ResourceShowUsage(command.Lister, version.CheckerMixin):
    """Show resource usages for a project (and optionally user) per class.

    Gives a report of usage information for resources associated with the
    project identified by the ``project_id`` argument and user identified by
    the ``--user-id`` option.

    This command requires at least ``--os-placement-api-version 1.9``.

    """

    def get_parser(self, prog_name):
        parser = super(ResourceShowUsage, self).get_parser(prog_name)

        parser.add_argument(
            'project_id',
            metavar='<project-uuid>',
            help='ID of the project.'
        )
        parser.add_argument(
            '--user-id',
            metavar='<user-uuid>',
            help='ID of the user.'
        )

        return parser

    @version.check(version.ge('1.9'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = USAGES_URL
        params = {'project_id': parsed_args.project_id}
        if parsed_args.user_id:
            params['user_id'] = parsed_args.user_id
        per_class = http.request(
            'GET', url, params=params).json()['usages']

        usages = [{'resource_class': k, 'usage': v}
                  for k, v in per_class.items()]
        rows = (utils.get_dict_properties(u, FIELDS) for u in usages)
        return FIELDS, rows


class ShowUsage(command.Lister):
    """Show resource usages per class for a given resource provider."""

    def get_parser(self, prog_name):
        parser = super(ShowUsage, self).get_parser(prog_name)

        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the resource provider'
        )

        return parser

    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = BASE_URL.format(uuid=parsed_args.uuid)
        per_class = http.request('GET', url).json()['usages']

        usages = [{'resource_class': k, 'usage': v}
                  for k, v in per_class.items()]
        rows = (utils.get_dict_properties(u, FIELDS) for u in usages)
        return FIELDS, rows



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from osc_placement.tests.functional import base


class TestAllocation(base.BaseTestCase):
    def setUp(self):
        super(TestAllocation, self).setUp()

        self.rp1 = self.resource_provider_create()
        self.resource_inventory_set(
            self.rp1['uuid'], 'VCPU=4', 'MEMORY_MB=1024')

    def test_allocation_show_not_found(self):
        consumer_uuid = str(uuid.uuid4())

        result = self.resource_allocation_show(consumer_uuid)
        self.assertEqual([], result)

    def test_allocation_create(self):
        consumer_uuid = str(uuid.uuid4())

        created_alloc = self.resource_allocation_set(
            consumer_uuid,
            ['rp={},VCPU=2'.format(self.rp1['uuid']),
             'rp={},MEMORY_MB=512'.format(self.rp1['uuid'])]
        )
        retrieved_alloc = self.resource_allocation_show(consumer_uuid)

        expected = [
            {'resource_provider': self.rp1['uuid'],
             'generation': 2,
             'resources': {'VCPU': 2, 'MEMORY_MB': 512}}
        ]
        self.assertEqual(expected, created_alloc)
        self.assertEqual(expected, retrieved_alloc)

        # Test that specifying --project-id and --user-id before microversion
        # 1.8 does not result in an error but display a warning.
        output, warning = self.resource_allocation_set(
            consumer_uuid,
            ['rp={},VCPU=2'.format(self.rp1['uuid']),
             'rp={},MEMORY_MB=512'.format(self.rp1['uuid'])],
            project_id='fake-project', user_id='fake-user',
            may_print_to_stderr=True)
        expected = [
            {'resource_provider': self.rp1['uuid'],
             'generation': 3,
             'resources': {'VCPU': 2, 'MEMORY_MB': 512}}
        ]
        self.assertEqual(expected, output)
        self.assertIn(
            '--project-id and --user-id options do not affect allocation for '
            '--os-placement-api-version less than 1.8', warning)

        # Test that specifying --consumer-type before microversion 1.38 does
        # not result in an error but display a warning.
        output, warning = self.resource_allocation_set(
            consumer_uuid,
            ['rp={},VCPU=2'.format(self.rp1['uuid']),
             'rp={},MEMORY_MB=512'.format(self.rp1['uuid'])],
            consumer_type='fake-type', may_print_to_stderr=True)
        expected = [
            {'resource_provider': self.rp1['uuid'],
             'generation': 4,
             'resources': {'VCPU': 2, 'MEMORY_MB': 512}}
        ]
        self.assertEqual(expected, output)
        self.assertIn(
            '--consumer-type option does not affect allocation for '
            '--os-placement-api-version less than 1.38', warning)

    def test_allocation_create_empty(self):
        consumer_uuid = str(uuid.uuid4())

        exc = self.assertRaises(base.CommandException,
                                self.resource_allocation_set,
                                consumer_uuid, [])
        self.assertIn('At least one resource allocation must be specified',
                      str(exc))

    def test_allocation_delete(self):
        consumer_uuid = str(uuid.uuid4())

        self.resource_allocation_set(
            consumer_uuid,
            ['rp={},VCPU=2'.format(self.rp1['uuid']),
             'rp={},MEMORY_MB=512'.format(self.rp1['uuid'])]
        )
        self.assertTrue(self.resource_allocation_show(consumer_uuid))

        self.resource_allocation_delete(consumer_uuid)
        self.assertEqual([], self.resource_allocation_show(consumer_uuid))

    def test_allocation_delete_not_found(self):
        consumer_uuid = str(uuid.uuid4())

        msg = "No allocations for consumer '{}'".format(consumer_uuid)
        exc = self.assertRaises(base.CommandException,
                                self.resource_allocation_delete, consumer_uuid)
        self.assertIn(msg, str(exc))


class TestAllocation18(base.BaseTestCase):
    VERSION = '1.8'

    def test_allocation_create(self):
        consumer_uuid = str(uuid.uuid4())
        project_id = str(uuid.uuid4())
        user_id = str(uuid.uuid4())

        rp1 = self.resource_provider_create()
        self.resource_inventory_set(
            rp1['uuid'],
            'VCPU=4',
            'VCPU:max_unit=4',
            'MEMORY_MB=1024',
            'MEMORY_MB:max_unit=1024')
        created_alloc = self.resource_allocation_set(
            consumer_uuid,
            ['rp={},VCPU=2'.format(rp1['uuid']),
             'rp={},MEMORY_MB=512'.format(rp1['uuid'])],
            project_id=project_id, user_id=user_id
        )
        retrieved_alloc = self.resource_allocation_show(consumer_uuid)

        expected = [
            {'resource_provider': rp1['uuid'],
             'generation': 2,
             'resources': {'VCPU': 2, 'MEMORY_MB': 512}}
        ]
        self.assertEqual(expected, created_alloc)
        self.assertEqual(expected, retrieved_alloc)


class TestAllocation112(base.BaseTestCase):
    VERSION = '1.12'

    def setUp(self):
        super(TestAllocation112, self).setUp()

        self.rp1 = self.resource_provider_create()
        self.resource_inventory_set(
            self.rp1['uuid'], 'VCPU=4', 'MEMORY_MB=1024')

    def test_allocation_update(self):
        consumer_uuid = str(uuid.uuid4())
        project_uuid = str(uuid.uuid4())
        user_uuid = str(uuid.uuid4())

        created_alloc = self.resource_allocation_set(
            consumer_uuid,
            ['rp={},VCPU=2'.format(self.rp1['uuid']),
             'rp={},MEMORY_MB=512'.format(self.rp1['uuid'])],
            project_id=project_uuid, user_id=user_uuid
        )
        retrieved_alloc = self.resource_allocation_show(consumer_uuid)

        expected = [
            {'resource_provider': self.rp1['uuid'],
             'generation': 2,
             'project_id': project_uuid,
             'user_id': user_uuid,
             'resources': {'VCPU': 2, 'MEMORY_MB': 512}}
        ]
        self.assertEqual(expected, created_alloc)
        self.assertEqual(expected, retrieved_alloc)

    def test_allocation_update_to_empty(self):
        consumer_uuid = str(uuid.uuid4())
        project_uuid = str(uuid.uuid4())
        user_uuid = str(uuid.uuid4())

        self.resource_allocation_set(
            consumer_uuid,
            ['rp={},VCPU=2'.format(self.rp1['uuid'])],
            project_id=project_uuid,
            user_id=user_uuid,
        )

        result = self.resource_allocation_unset(
            consumer_uuid, columns=("resources",))

        self.assertEqual([], result)

    def test_allocation_show_empty(self):
        alloc = self.resource_allocation_show(
            str(uuid.uuid4()), columns=('resources',))
        self.assertEqual([], alloc)


class TestAllocation128(TestAllocation112):
    """Tests allocation set command with --os-placement-api-version 1.28.

    The 1.28 microversion adds the consumer_generation parameter to the
    GET and PUT /allocations/{consumer_id} APIs.
    """
    VERSION = '1.28'

    def test_allocation_update(self):
        consumer_uuid = str(uuid.uuid4())
        project_uuid = str(uuid.uuid4())
        user_uuid = str(uuid.uuid4())
        # First create the initial set of allocations using rp1.
        created_alloc = self.resource_allocation_set(
            consumer_uuid,
            ['rp={},VCPU=2'.format(self.rp1['uuid']),
             'rp={},MEMORY_MB=512'.format(self.rp1['uuid'])],
            project_id=project_uuid, user_id=user_uuid
        )
        retrieved_alloc = self.resource_allocation_show(consumer_uuid)

        expected = [
            {'resource_provider': self.rp1['uuid'],
             'generation': 2,
             'project_id': project_uuid,
             'user_id': user_uuid,
             'resources': {'VCPU': 2, 'MEMORY_MB': 512}}
        ]
        self.assertEqual(expected, created_alloc)
        self.assertEqual(expected, retrieved_alloc)
        # Now update the allocations which should use the consumer generation.
        updated_alloc = self.resource_allocation_set(
            consumer_uuid,
            ['rp={},VCPU=4'.format(self.rp1['uuid']),
             'rp={},MEMORY_MB=1024'.format(self.rp1['uuid'])],
            project_id=project_uuid, user_id=user_uuid
        )
        expected[0].update({
            'generation': expected[0]['generation'] + 1,
            'resources': {'VCPU': 4, 'MEMORY_MB': 1024}
        })
        self.assertEqual(expected, updated_alloc)


class TestAllocation138(TestAllocation128):
    """Tests allocation set command with --os-placement-api-version 1.38.

    The 1.38 microversion adds the consumer_type parameter to the
    GET and PUT /allocations/{consumer_id} APIs
    """
    VERSION = '1.38'

    def test_allocation_update(self):
        consumer_uuid = str(uuid.uuid4())
        project_uuid = str(uuid.uuid4())
        user_uuid = str(uuid.uuid4())
        # First create the initial set of allocations using rp1.
        created_alloc = self.resource_allocation_set(
            consumer_uuid,
            ['rp={},VCPU=2'.format(self.rp1['uuid']),
             'rp={},MEMORY_MB=512'.format(self.rp1['uuid'])],
            project_id=project_uuid, user_id=user_uuid,
            consumer_type='INSTANCE'
        )
        retrieved_alloc = self.resource_allocation_show(consumer_uuid)

        expected = [
            {'resource_provider': self.rp1['uuid'],
             'generation': 2,
             'project_id': project_uuid,
             'user_id': user_uuid,
             'resources': {'VCPU': 2, 'MEMORY_MB': 512},
             'consumer_type': 'INSTANCE'}
        ]
        self.assertEqual(expected, created_alloc)
        self.assertEqual(expected, retrieved_alloc)
        # Now update the allocations which should use the consumer generation.
        updated_alloc = self.resource_allocation_set(
            consumer_uuid,
            ['rp={},VCPU=4'.format(self.rp1['uuid']),
             'rp={},MEMORY_MB=1024'.format(self.rp1['uuid'])],
            project_id=project_uuid, user_id=user_uuid,
            consumer_type='MIGRATION'
        )
        expected[0].update({
            'generation': expected[0]['generation'] + 1,
            'resources': {'VCPU': 4, 'MEMORY_MB': 1024},
            'consumer_type': 'MIGRATION'
        })
        self.assertEqual(expected, updated_alloc)

    def test_allocation_update_to_empty(self):
        consumer_uuid = str(uuid.uuid4())
        project_uuid = str(uuid.uuid4())
        user_uuid = str(uuid.uuid4())

        self.resource_allocation_set(
            consumer_uuid,
            ['rp={},VCPU=2'.format(self.rp1['uuid'])],
            project_id=project_uuid,
            user_id=user_uuid,
            consumer_type="INSTANCE",
        )

        result = self.resource_allocation_unset(
            consumer_uuid, columns=('resources', 'consumer_type'))
        self.assertEqual([], result)


class TestAllocationUnsetOldVersion(base.BaseTestCase):

    def test_invalid_version(self):
        """Negative test to ensure the unset command requires >= 1.12."""
        consumer_uuid = str(uuid.uuid4())
        self.assertCommandFailed('requires at least version 1.12',
                                 self.resource_allocation_unset, consumer_uuid)


class TestAllocationUnset112(base.BaseTestCase):
    VERSION = '1.12'

    def setUp(self):
        super(TestAllocationUnset112, self).setUp()
        # Create four providers with inventory.
        self.rp1 = self.resource_provider_create()
        self.rp2 = self.resource_provider_create()
        self.rp3 = self.resource_provider_create()
        self.rp4 = self.resource_provider_create()

        self.resource_inventory_set(
            self.rp1['uuid'], 'VCPU=4', 'MEMORY_MB=1024')
        self.resource_inventory_set(
            self.rp2['uuid'], 'VGPU=1')
        self.resource_inventory_set(
            self.rp3['uuid'], 'VCPU=4', 'MEMORY_MB=1024', 'VGPU=1')
        self.resource_inventory_set(
            self.rp4['uuid'], 'VCPU=4', 'MEMORY_MB=1024', 'VGPU=1')

        self.consumer_uuid1 = str(uuid.uuid4())
        self.consumer_uuid2 = str(uuid.uuid4())
        self.project_uuid = str(uuid.uuid4())
        self.user_uuid = str(uuid.uuid4())

        # Create allocations against rp1 and rp2 for consumer1.
        self.resource_allocation_set(
            self.consumer_uuid1,
            ['rp={},VCPU=2'.format(self.rp1['uuid']),
             'rp={},MEMORY_MB=512'.format(self.rp1['uuid']),
             'rp={},VGPU=1'.format(self.rp2['uuid'])],
            project_id=self.project_uuid, user_id=self.user_uuid)

        # Create allocations against rp3 and rp4 for consumer1.
        self.resource_allocation_set(
            self.consumer_uuid2,
            ['rp={},VCPU=1'.format(self.rp3['uuid']),
             'rp={},MEMORY_MB=256'.format(self.rp3['uuid']),
             'rp={},VGPU=1'.format(self.rp3['uuid']),
             'rp={},VCPU=1'.format(self.rp4['uuid']),
             'rp={},MEMORY_MB=256'.format(self.rp4['uuid'])],
            project_id=self.project_uuid, user_id=self.user_uuid)

    def test_allocation_unset_one_provider(self):
        """Tests removing allocations for one specific provider."""
        # Remove the allocation for rp1.
        updated_allocs = self.resource_allocation_unset(
            self.consumer_uuid1, provider=self.rp1['uuid'])
        expected = [
            {'resource_provider': self.rp2['uuid'],
             'generation': 3,
             'project_id': self.project_uuid,
             'user_id': self.user_uuid,
             'resources': {'VGPU': 1}}
        ]
        self.assertEqual(expected, updated_allocs)

    def test_allocation_unset_one_resource_class(self):
        """Tests removing allocations for resource classes."""
        updated_allocs = self.resource_allocation_unset(
            self.consumer_uuid2, resource_class=['MEMORY_MB'])
        expected = [
            {'resource_provider': self.rp3['uuid'],
             'generation': 3,
             'project_id': self.project_uuid,
             'user_id': self.user_uuid,
             'resources': {'VCPU': 1, 'VGPU': 1}},
            {'resource_provider': self.rp4['uuid'],
             'generation': 3,
             'project_id': self.project_uuid,
             'user_id': self.user_uuid,
             'resources': {'VCPU': 1}}
        ]
        self.assertEqual(expected, updated_allocs)

    def test_allocation_unset_resource_classes(self):
        """Tests removing allocations for resource classes."""
        updated_allocs = self.resource_allocation_unset(
            self.consumer_uuid2, resource_class=['VCPU', 'MEMORY_MB'])
        expected = [
            {'resource_provider': self.rp3['uuid'],
             'generation': 3,
             'project_id': self.project_uuid,
             'user_id': self.user_uuid,
             'resources': {'VGPU': 1}}
        ]
        self.assertEqual(expected, updated_allocs)

    def test_allocation_unset_provider_and_rc(self):
        """Tests removing allocations of resource classes for a provider ."""
        updated_allocs = self.resource_allocation_unset(
            self.consumer_uuid2, provider=self.rp3['uuid'],
            resource_class=['VCPU', 'MEMORY_MB'])
        expected = [
            {'resource_provider': self.rp3['uuid'],
             'generation': 3,
             'project_id': self.project_uuid,
             'user_id': self.user_uuid,
             'resources': {'VGPU': 1}},
            {'resource_provider': self.rp4['uuid'],
             'generation': 3,
             'project_id': self.project_uuid,
             'user_id': self.user_uuid,
             'resources': {'VCPU': 1, 'MEMORY_MB': 256}},
        ]
        self.assertEqual(expected, updated_allocs)

    def test_allocation_unset_remove_all_providers(self):
        """Tests removing all allocations by omitting the --provider option."""
        # For this test pass use_json=False to make sure we get nothing back
        # in the output since there are no more allocations.
        updated_allocs = self.resource_allocation_unset(
            self.consumer_uuid1, use_json=False)
        self.assertEqual('', updated_allocs.strip())


class TestAllocationUnset128(TestAllocationUnset112):
    """Tests allocation unset command with --os-placement-api-version 1.28.

    The 1.28 microversion adds the consumer_generation parameter to the
    GET and PUT /allocations/{consumer_id} APIs.
    """
    VERSION = '1.28'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import contextlib
import json
import logging

import keystoneauth1.exceptions.http as ks_exceptions
import osc_lib.exceptions as exceptions

from osc_placement import version


_http_error_to_exc = {
    cls.http_status: cls
    for cls in exceptions.ClientException.__subclasses__()
}


LOG = logging.getLogger(__name__)


@contextlib.contextmanager
def _wrap_http_exceptions():
    """Reraise osc-lib exceptions with detailed messages."""

    try:
        yield
    except ks_exceptions.HttpError as exc:
        if 400 <= exc.http_status < 500:
            detail = json.loads(exc.response.content)['errors'][0]['detail']
            msg = detail.split('\n')[-1].strip()
            exc_class = _http_error_to_exc.get(exc.http_status,
                                               exceptions.CommandError)
            raise exc_class(exc.http_status, msg) from exc
        else:
            raise


class SessionClient(object):
    def __init__(self, session, ks_filter, api_version='1.0'):
        self.session = session
        self.ks_filter = ks_filter
        self.negotiate_api_version(api_version)

    def request(self, method, url, **kwargs):
        version = kwargs.pop('version', None)
        api_version = (self.ks_filter['service_type'] + ' '
                       + (version or self.api_version))
        headers = kwargs.pop('headers', {})
        headers.setdefault('OpenStack-API-Version', api_version)
        headers.setdefault('Accept', 'application/json')

        with _wrap_http_exceptions():
            return self.session.request(url, method,
                                        headers=headers,
                                        endpoint_filter=self.ks_filter,
                                        **kwargs)

    def negotiate_api_version(self, api_version):
        """Set api_version to self.

        If negotiate version (only majorversion) is given, talk to server to
        pick up max microversion supported both by client and by server.
        """
        if api_version not in version.NEGOTIATE_VERSIONS:
            self.api_version = api_version
            return
        client_ver = version.MAX_VERSION_NO_GAP
        self.api_version = client_ver
        resp = self.request('GET', '/', raise_exc=False)
        if resp.status_code == 406:
            server_ver = resp.json()['errors'][0]['max_version']
            self.api_version = server_ver
            LOG.debug('Microversion %s not supported in server. '
                      'Falling back to microversion %s',
                      client_ver, server_ver)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from urllib import parse as urlparse


def encode(value, encoding='utf-8'):
    """Return a byte repr of a string for a given encoding.

    Byte strings and values of other types are returned as is.

    """

    if isinstance(value, str):
        return value.encode(encoding)
    else:
        return value


def url_with_filters(url, filters=None):
    """Add a percent-encoded string of filters (a dict) to a base url."""

    if filters:
        filters = [(encode(k), encode(v)) for k, v in filters.items()]

        urlencoded_filters = urlparse.urlencode(filters)
        url = urlparse.urljoin(url, '?' + urlencoded_filters)

    return url


def get_required_query_param_from_args(required_traits, forbidden_traits):
    # Iterate the required params and collect OR groups and simple
    # AND traits separately. Each OR group needs a separate query param
    # while the AND traits and forbidden traits can be collated to a single
    # query param
    required_query_params = []
    and_traits = []
    for required in required_traits:
        if ',' in required:
            required_query_params.append('in:' + required)
        else:
            and_traits.append(required)
    # We need an extra required query param for the and_traits and the
    # forbidden traits
    and_query = ','.join(and_traits + forbidden_traits)
    if and_query:
        required_query_params.append(and_query)
    return required_query_params



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from osc_placement.tests.functional import base


TRAIT = 'CUSTOM_FAKE_HW_GPU_CLASS_{}'.format(
    str(uuid.uuid4()).replace('-', '').upper())


class TestTrait(base.BaseTestCase):
    VERSION = '1.6'

    def test_list_traits(self):
        self.assertTrue(len(self.trait_list()) > 0)

    def test_list_associated_traits(self):
        self.trait_create(TRAIT)
        rp = self.resource_provider_create()
        self.resource_provider_trait_set(rp['uuid'], TRAIT)
        self.assertIn(TRAIT,
                      {t['name'] for t in self.trait_list(associated=True)})

    def test_list_traits_startswith(self):
        self.trait_create(TRAIT)
        rp = self.resource_provider_create()
        self.resource_provider_trait_set(rp['uuid'], TRAIT)
        traits = {t['name'] for t in self.trait_list(
            name='startswith:' + TRAIT)}
        self.assertEqual(1, len(traits))
        self.assertIn(TRAIT, traits)

    def test_list_traits_startswith_unknown_trait(self):
        traits = {t['name'] for t in self.trait_list(
            name='startswith:CUSTOM_FOO')}
        self.assertEqual(0, len(traits))

    def test_list_traits_in(self):
        self.trait_create(TRAIT)
        rp = self.resource_provider_create()
        self.resource_provider_trait_set(rp['uuid'], TRAIT)
        traits = {t['name'] for t in self.trait_list(
            name='in:' + TRAIT)}
        self.assertEqual(1, len(traits))
        self.assertIn(TRAIT, traits)

    def test_list_traits_in_unknown_trait(self):
        traits = {t['name'] for t in self.trait_list(name='in:CUSTOM_FOO')}
        self.assertEqual(0, len(traits))

    def test_show_trait(self):
        self.trait_create(TRAIT)
        self.assertEqual({'name': TRAIT}, self.trait_show(TRAIT))

    def test_fail_show_unknown_trait(self):
        self.assertCommandFailed('HTTP 404', self.trait_show, 'UNKNOWN')

    def test_set_multiple_traits(self):
        self.trait_create(TRAIT + '1')
        self.trait_create(TRAIT + '2')
        rp = self.resource_provider_create()
        self.resource_provider_trait_set(rp['uuid'], TRAIT + '1', TRAIT + '2')
        traits = {t['name'] for t in self.resource_provider_trait_list(
            rp['uuid'])}
        self.assertEqual(2, len(traits))

    def test_set_known_and_unknown_traits(self):
        self.trait_create(TRAIT)
        rp = self.resource_provider_create()
        self.assertCommandFailed(
            'No such trait',
            self.resource_provider_trait_set, rp['uuid'], TRAIT, TRAIT + '1')
        self.assertEqual([], self.resource_provider_trait_list(rp['uuid']))

    def test_delete_traits_from_provider(self):
        self.trait_create(TRAIT)
        rp = self.resource_provider_create()
        self.resource_provider_trait_set(rp['uuid'], TRAIT)
        traits = {t['name'] for t in self.resource_provider_trait_list(
            rp['uuid'])}
        self.assertEqual(1, len(traits))
        self.assertIn(TRAIT, traits)
        self.resource_provider_trait_delete(rp['uuid'])
        traits = {t['name'] for t in self.resource_provider_trait_list(
            rp['uuid'])}
        self.assertEqual(0, len(traits))

    def test_delete_trait(self):
        self.trait_create(TRAIT)
        self.trait_delete(TRAIT)
        self.assertCommandFailed('HTTP 404', self.trait_show, TRAIT)

    def test_fail_rp_trait_list_unknown_uuid(self):
        self.assertCommandFailed(
            'No resource provider', self.resource_provider_trait_list, 123)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from osc_lib.command import command
from osc_lib import exceptions
from osc_placement import version


BASE_URL = '/resource_providers/{uuid}/aggregates'
FIELDS = ('uuid',)


class SetAggregate(command.Lister, version.CheckerMixin):

    """Associate a list of aggregates with the resource provider.

    Each request cleans up previously associated resource provider
    aggregates entirely and sets the new ones. Passing empty aggregate
    UUID list will remove all associations with aggregates for the
    particular resource provider.

    This command requires at least ``--os-placement-api-version 1.1``.
    """

    def get_parser(self, prog_name):
        parser = super(SetAggregate, self).get_parser(prog_name)

        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the resource provider'
        )

        parser.add_argument(
            '--aggregate',
            metavar='<aggregate_uuid>',
            help='UUID of the aggregate. Specify multiple times to associate '
                 'a resource provider with multiple aggregates.',
            action='append',
            default=[]
        )

        parser.add_argument(
            '--generation',
            metavar='<resource_provider_generation>',
            type=int,
            help='The generation of resource provider. Must match the server-'
                 'side generation of the resource provider or the operation '
                 'will fail.\n\n'
                 'This param requires at least '
                 '``--os-placement-api-version 1.19``.'
        )

        return parser

    @version.check(version.ge('1.1'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = BASE_URL.format(uuid=parsed_args.uuid)
        aggregate = parsed_args.aggregate
        generation = None
        if 'generation' in parsed_args and parsed_args.generation is not None:
            self.check_version(version.ge('1.19'))
            generation = parsed_args.generation

        if self.compare_version(version.lt('1.19')):
            resp = http.request('PUT', url, json=aggregate).json()
        # Microversion 1.19 and beyond a generation argument is
        # required to write aggregates.
        elif generation is not None:
            data = {'aggregates': aggregate,
                    'resource_provider_generation': generation}
            resp = http.request('PUT', url, json=data).json()
        else:
            raise exceptions.CommandError(
                'A generation must be specified.')

        return FIELDS, [[r] for r in resp['aggregates']]


class ListAggregate(command.Lister):

    """List resource provider aggregates.

    This command requires at least ``--os-placement-api-version 1.1``.
    """

    def get_parser(self, prog_name):
        parser = super(ListAggregate, self).get_parser(prog_name)

        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the resource provider'
        )

        return parser

    @version.check(version.ge('1.1'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = BASE_URL.format(uuid=parsed_args.uuid)
        resp = http.request('GET', url).json()
        return FIELDS, [[r] for r in resp['aggregates']]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import operator
import uuid

from osc_placement.tests.functional import base


class TestUsage(base.BaseTestCase):
    def test_usage_show(self):
        consumer_uuid = str(uuid.uuid4())
        rp = self.resource_provider_create()
        self.resource_inventory_set(
            rp['uuid'],
            'VCPU=4',
            'VCPU:max_unit=4',
            'MEMORY_MB=1024',
            'MEMORY_MB:max_unit=1024')

        self.assertEqual([{'resource_class': 'MEMORY_MB', 'usage': 0},
                          {'resource_class': 'VCPU', 'usage': 0}],
                         sorted(self.resource_provider_show_usage(rp['uuid']),
                                key=operator.itemgetter('resource_class')))

        self.resource_allocation_set(
            consumer_uuid,
            ['rp={},VCPU=2'.format(rp['uuid']),
             'rp={},MEMORY_MB=512'.format(rp['uuid'])]
        )
        self.assertEqual([{'resource_class': 'MEMORY_MB', 'usage': 512},
                          {'resource_class': 'VCPU', 'usage': 2}],
                         sorted(self.resource_provider_show_usage(rp['uuid']),
                                key=operator.itemgetter('resource_class')))

    def test_usage_not_found(self):
        rp_uuid = str(uuid.uuid4())

        exc = self.assertRaises(base.CommandException,
                                self.resource_provider_show_usage,
                                rp_uuid)
        self.assertIn(
            'No resource provider with uuid {} found'.format(rp_uuid),
            str(exc)
        )

    def test_usage_empty(self):
        rp = self.resource_provider_create()

        self.assertEqual([], self.resource_provider_show_usage(rp['uuid']))


class TestResourceUsage(base.BaseTestCase):
    VERSION = '1.9'

    def test_usage_by_project_id_user_id(self):
        c1 = str(uuid.uuid4())
        c2 = str(uuid.uuid4())
        c3 = str(uuid.uuid4())
        p1 = str(uuid.uuid4())
        p2 = str(uuid.uuid4())
        u1 = str(uuid.uuid4())
        u2 = str(uuid.uuid4())

        rp = self.resource_provider_create()
        self.resource_inventory_set(rp['uuid'], 'VCPU=16')
        self.resource_allocation_set(
            c1, ['rp={},VCPU=2'.format(rp['uuid'])], project_id=p1, user_id=u1)
        self.resource_allocation_set(
            c2, ['rp={},VCPU=4'.format(rp['uuid'])], project_id=p2, user_id=u1)
        self.resource_allocation_set(
            c3, ['rp={},VCPU=6'.format(rp['uuid'])], project_id=p1, user_id=u2)

        # Show usage on the resource provider for all consumers.
        self.assertEqual(
            12, self.resource_provider_show_usage(uuid=rp['uuid'])[0]['usage'])
        # Show usage for project p1.
        self.assertEqual(
            8, self.resource_show_usage(project_id=p1)[0]['usage'])
        # Show usage for project p1 and user u1.
        self.assertEqual(
            2, self.resource_show_usage(
                project_id=p1, user_id=u1)[0]['usage'])
        # Show usage for project p2.
        self.assertEqual(
            4, self.resource_show_usage(project_id=p2)[0]['usage'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import collections

from osc_lib.command import command
from osc_lib import exceptions

from osc_placement.resources import common
from osc_placement import version


BASE_URL = '/allocation_candidates'


class GroupAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        group, = values
        namespace._current_group = group
        groups = namespace.__dict__.setdefault('groups', {})
        groups[group] = collections.defaultdict(list)


class AppendToGroup(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if getattr(namespace, '_current_group', None) is None:
            groups = namespace.__dict__.setdefault('groups', {})
            namespace._current_group = ''
            groups[''] = collections.defaultdict(list)
        namespace.groups[namespace._current_group][self.dest].append(values)


class ListAllocationCandidate(command.Lister, version.CheckerMixin):

    """List allocation candidates.

    Returns a representation of a collection of allocation requests and
    resource provider summaries. Each allocation request has information
    to issue an ``openstack resource provider allocation set`` request to claim
    resources against a related set of resource providers.

    As several allocation requests are available its necessary to select one.
    To make a decision, resource provider summaries are provided with the
    inventory/capacity information.

    For example::

      $ export OS_PLACEMENT_API_VERSION=1.10
      $ openstack allocation candidate list --resource VCPU=1
      +---+------------+-------------------------+-------------------------+
      | # | allocation | resource provider       | inventory used/capacity |
      +---+------------+-------------------------+-------------------------+
      | 1 | VCPU=1     | 66bcaca9-9263-45b1-a569 | VCPU=0/128              |
      |   |            | -ea708ff7a968           |                         |
      +---+------------+-------------------------+-------------------------+

    In this case, the user is looking for resource providers that can have
    capacity to allocate 1 ``VCPU`` resource class. There is one resource
    provider that can serve that allocation request and that resource providers
    current ``VCPU`` inventory used is 0 and available capacity is 128.

    This command requires at least ``--os-placement-api-version 1.10``.
    """

    def get_parser(self, prog_name):
        parser = super(ListAllocationCandidate, self).get_parser(prog_name)

        parser.add_argument(
            '--resource',
            metavar='<resource_class>=<value>',
            dest='resources',
            action=AppendToGroup,
            help='String indicating an amount of resource of a specified '
                 'class that providers in each allocation request must '
                 'collectively have the capacity and availability to serve. '
                 'Can be specified multiple times per resource class. '
                 'For example: '
                 '``--resource VCPU=4 --resource DISK_GB=64 '
                 '--resource MEMORY_MB=2048``'
        )
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            help='A positive integer to limit '
                 'the maximum number of allocation candidates. '
                 'This option requires at least '
                 '``--os-placement-api-version 1.16``.'
        )
        parser.add_argument(
            '--required',
            metavar='<required>',
            action=AppendToGroup,
            help='A required trait. May be repeated. Allocation candidates '
                 'must collectively contain all of the required traits. '
                 'This option requires at least '
                 '``--os-placement-api-version 1.17``. '
                 'Since ``--os-placement-api-version 1.39`` the value of '
                 'this parameter can be a comma separated list of trait names '
                 'to express OR relationship between those traits.'
        )
        parser.add_argument(
            '--forbidden',
            metavar='<forbidden>',
            action=AppendToGroup,
            help='A forbidden trait. May be repeated. Returned allocation '
                 'candidates must not contain any of the specified traits. '
                 'This option requires at least '
                 '``--os-placement-api-version 1.22``.'
        )
        # NOTE(tetsuro): --aggregate-uuid is deprecated in Jan 2020 in 1.x
        # release. Do not remove before Jan 2021 and a 2.x release.
        aggregate_group = parser.add_mutually_exclusive_group()
        aggregate_group.add_argument(
            "--member-of",
            action=AppendToGroup,
            metavar='<member_of>',
            help='A list of comma-separated UUIDs of the resource provider '
                 'aggregates. The returned allocation candidates must be '
                 'associated with at least one of the aggregates identified '
                 'by uuid. This param requires at least '
                 '``--os-placement-api-version 1.21`` and can be repeated to '
                 'add(restrict) the condition with '
                 '``--os-placement-api-version 1.24`` or greater. '
                 'For example, to get candidates in either of agg1 or agg2 '
                 'and definitely in agg3, specify:\n\n'
                 '``--member_of <agg1>,<agg2> --member_of <agg3>``'
        )
        aggregate_group.add_argument(
            '--aggregate-uuid',
            action=AppendToGroup,
            metavar='<aggregate_uuid>',
            help=argparse.SUPPRESS
        )
        parser.add_argument(
            '--group',
            action=GroupAction,
            metavar='<group>',
            help='An integer to group granular requests. If specified, '
                 'following given options of resources, required/forbidden '
                 'traits, and aggregate are associated to that group and will '
                 'be satisfied by the same resource provider in the response. '
                 'Can be repeated to get candidates from multiple resource '
                 'providers in the same resource provider tree. '
                 'For example, ``--group 1 --resource VCPU=3 --required '
                 'HW_CPU_X86_AVX --group 2 --resource VCPU=2 --required '
                 'HW_CPU_X86_SSE`` will provide candidates where three VCPUs '
                 'comes from a provider with ``HW_CPU_X86_AVX`` trait and '
                 'two VCPUs from a provider with ``HW_CPU_X86_SSE`` trait. '
                 'This option requires at least '
                 '``--os-placement-api-version 1.25`` or greater, but to have '
                 'placement server be aware of resource provider tree, use '
                 '``--os-placement-api-version 1.29`` or greater.'
        )
        parser.add_argument(
            '--group-policy',
            choices=['none', 'isolate'],
            default='none',
            metavar='<group_policy>',
            help='This indicates how the groups should interact when multiple '
                 'groups are supplied. With group_policy=none (default), '
                 'separate groups may or may not be satisfied by the same '
                 'provider. With group_policy=isolate, numbered groups are '
                 'guaranteed to be satisfied by different providers.'
        )

        return parser

    @version.check(version.ge('1.10'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        params = {}
        if 'groups' not in parsed_args:
            raise exceptions.CommandError(
                'At least one --resource must be specified.')

        if 'limit' in parsed_args and parsed_args.limit:
            # Fail if --limit but not high enough microversion.
            self.check_version(version.ge('1.16'))
            params['limit'] = int(parsed_args.limit)

        if any(parsed_args.groups):
            self.check_version(version.ge('1.25'))
            params['group_policy'] = parsed_args.group_policy

        for suffix, group in parsed_args.groups.items():
            def _get_key(name):
                return name + suffix

            if 'resources' not in group:
                raise exceptions.CommandError(
                    '--resources should be provided in group %s', suffix)
            for resource in group['resources']:
                if not len(resource.split('=')) == 2:
                    raise exceptions.CommandError(
                        'Arguments to --resource must be of form '
                        '<resource_class>=<value>')

            params[_get_key('resources')] = ','.join(
                resource.replace('=', ':') for resource in group['resources'])

            # We need to handle required and forbidden together as they all
            # end up in the same query param on the API.
            # First just check that the requested feature is aligned with the
            # request microversion
            required_traits = []
            if 'required' in group and group['required']:
                # Fail if --required but not high enough microversion.
                self.check_version(version.ge('1.17'))
                if any(',' in required for required in group['required']):
                    self.check_version(version.ge('1.39'))
                required_traits = group['required']

            forbidden_traits = []
            if 'forbidden' in group and group['forbidden']:
                self.check_version(version.ge('1.22'))
                forbidden_traits = ['!' + f for f in group['forbidden']]

            # Then collect the required query params containing both required
            # and forbidden traits
            params[_get_key('required')] = (
                common.get_required_query_param_from_args(
                    required_traits, forbidden_traits)
            )

            if 'aggregate_uuid' in group and group['aggregate_uuid']:
                # Fail if --aggregate_uuid but not high enough microversion.
                self.check_version(version.ge('1.21'))
                self.deprecated_option_warning(
                    "--aggregate-uuid", "--member-of")
                params[_get_key('member_of')] = 'in:' + ','.join(
                    group['aggregate_uuid'])
            if 'member_of' in group and group['member_of']:
                # Fail if --member-of but not high enough microversion.
                self.check_version(version.ge('1.21'))
                params[_get_key('member_of')] = [
                    'in:' + aggs for aggs in group['member_of']]

        resp = http.request('GET', BASE_URL, params=params).json()

        rp_resources = {}
        include_traits = self.compare_version(version.ge('1.17'))
        if include_traits:
            rp_traits = {}
        for rp_uuid, resources in resp['provider_summaries'].items():
            rp_resources[rp_uuid] = ','.join(
                '%s=%s/%s' % (rc, value['used'], value['capacity'])
                for rc, value in resources['resources'].items())
            if include_traits:
                rp_traits[rp_uuid] = ','.join(resources['traits'])

        rows = []
        if self.compare_version(version.ge('1.12')):
            for i, allocation_req in enumerate(resp['allocation_requests']):
                for rp, resources in allocation_req['allocations'].items():
                    req = ','.join(
                        '%s=%s' % (rc, value)
                        for rc, value in resources['resources'].items())
                    if include_traits:
                        row = [i + 1, req, rp, rp_resources[rp], rp_traits[rp]]
                    else:
                        row = [i + 1, req, rp, rp_resources[rp]]
                    rows.append(row)
        else:
            for i, allocation_req in enumerate(resp['allocation_requests']):
                for allocation in allocation_req['allocations']:
                    rp = allocation['resource_provider']['uuid']
                    req = ','.join(
                        '%s=%s' % (rc, value)
                        for rc, value in allocation['resources'].items())
                    rows.append([i + 1, req, rp, rp_resources[rp]])

        fields = ('#', 'allocation', 'resource provider',
                  'inventory used/capacity')
        if include_traits:
            fields += ('traits',)

        return fields, rows
